//
// SPDX-License-Identifier: BSD-3-Clause
// Copyright (c) Weta Digital, Ltd and Contributors to the OpenEXR Project.
//

#ifndef TESTCOMPOSITEDEEPSCANLINE_H_
#define TESTCOMPOSITEDEEPSCANLINE_H_

#include <string>

void testCompositeDeepScanLine (const std::string & tempDir);

#endif /* TESTCOMPOSITEDEEPSCANLINE_H_ */
